@@ -1,31 +1,34 @@
 from statefun_tasks.context import TaskContext
 from statefun_tasks.messages_pb2 import ChildPipeline, TaskRequest, TaskResult, TaskException, Pipeline, TaskActionRequest, TaskAction, TaskStatus
 from statefun_tasks.serialisation import DefaultSerialiser
-from statefun_tasks.types import Task, Group, TaskCancelledException
+from statefun_tasks.storage import StorageBackend
+from statefun_tasks.types import Task, Group, TaskCancelledException, PipelineInProgress
 from statefun_tasks.type_helpers import _create_task_exception
 from statefun_tasks.pipeline_impl.handlers import BeginPipelineHandler, ContinuePipelineHandler, EndPipelineHandler, CancelPipelineHandler
 from statefun_tasks.pipeline_impl.helpers import PipelineGraph, DeferredTaskSubmitter
 from statefun_tasks.events import EventHandlers
+from statefun_tasks.utils import _gen_id
 from google.protobuf.any_pb2 import Any
 from typing import Union
 
 
 class _Pipeline(object):
-    def __init__(self, pipeline: list, serialiser=None, is_fruitful=True, events: EventHandlers=None):
+    def __init__(self, pipeline: list, serialiser=None, is_fruitful=True, events: EventHandlers=None, storage: StorageBackend=None):
         self._pipeline = pipeline
         self._serialiser = serialiser or DefaultSerialiser()
         self._is_fruitful = is_fruitful
         self._events = events or EventHandlers()
+        self._storage = storage
 
         self._handlers = [
-            BeginPipelineHandler(self._pipeline, self._serialiser),
-            ContinuePipelineHandler(self._pipeline, self._serialiser),
-            CancelPipelineHandler(self._pipeline, self._serialiser),
-            EndPipelineHandler(self._pipeline, self._serialiser)
+            BeginPipelineHandler(self._pipeline, self._serialiser, self._storage),
+            ContinuePipelineHandler(self._pipeline, self._serialiser, self._storage),
+            CancelPipelineHandler(self._pipeline, self._serialiser, self._storage),
+            EndPipelineHandler(self._pipeline, self._serialiser, self._storage)
         ]
 
         self._graph = PipelineGraph(self._pipeline)
-        self._submitter = DeferredTaskSubmitter(self._graph, self._serialiser)
+        self._submitter = DeferredTaskSubmitter(self._graph, self._serialiser, self._storage)
 
     @property
     def events(self) -> EventHandlers:
@@ -43,7 +46,7 @@ class _Pipeline(object):
         return pipeline
 
     @staticmethod
-    def from_proto(pipeline_proto: Pipeline, serialiser, events):
+    def from_proto(pipeline_proto: Pipeline, serialiser, events, storage):
         pipeline = []
 
         for proto in pipeline_proto.entries:
@@ -52,14 +55,14 @@ class _Pipeline(object):
             elif proto.HasField('group_entry'):
                 pipeline.append(Group.from_proto(proto))
 
-        return _Pipeline(pipeline, serialiser=serialiser, events=events)
+        return _Pipeline(pipeline, serialiser=serialiser, events=events, storage=storage)
 
-    def handle_message(self, context: TaskContext, message: Union[TaskRequest, TaskResult, TaskException], task_state: Any=None) -> bool:
+    async def handle_message(self, context: TaskContext, message: Union[TaskRequest, TaskResult, TaskException], state: Any=None) -> bool:
         handled = False
 
         for handler in self._handlers:
             if handler.can_handle_message(context, message):
-                should_continue, message = handler.handle_message(context, message, pipeline=self, task_state=task_state)
+                should_continue, message = await handler.handle_message(context, message, pipeline=self, state=state)
                 handled = True
 
                 if not should_continue:
@@ -88,33 +91,38 @@ class _Pipeline(object):
     def status(self, context: TaskContext):
         return context.pipeline_state.status
 
-    def pause(self, context: TaskContext):
+    async def pause(self, context: TaskContext):
         if context.pipeline_state.status.value not in [TaskStatus.PENDING, TaskStatus.RUNNING, TaskStatus.PAUSED]:
             raise ValueError(f'Pipeline is not in a state that can be paused')
 
         context.pipeline_state.status.value = TaskStatus.Status.PAUSED
-        self.events.notify_pipeline_status_changed(context.pipeline_state.pipeline, context.pipeline_state.status.value)
+        self.events.notify_pipeline_status_changed(context, context.pipeline_state.pipeline, context.pipeline_state.status.value)
 
         # tell any child pipelines to pause
         for child_pipeline in context.pipeline_state.child_pipelines:
-            pause_action = TaskActionRequest(id=child_pipeline.id, action=TaskAction.PAUSE_PIPELINE)
+            pause_action = TaskActionRequest(id=child_pipeline.id, uid=_gen_id(), action=TaskAction.PAUSE_PIPELINE)
             context.send_message(child_pipeline.address, pause_action.id, pause_action)
 
-    def unpause(self, context: TaskContext):
+    async def unpause(self, context: TaskContext):
         if context.pipeline_state.status.value not in [TaskStatus.PENDING, TaskStatus.RUNNING, TaskStatus.PAUSED]:
             raise ValueError(f'Pipeline is not in a state that can be unpaused')
         
         context.pipeline_state.status.value = TaskStatus.Status.RUNNING
         self.events.notify_pipeline_status_changed(context, context.pipeline_state.pipeline, context.pipeline_state.status.value)
 
-        self._submitter.unpause_tasks(context)
+        try:
+            await self._submitter.unpause_tasks(context)
 
-        # tell any child pipelines to resume
-        for child_pipeline in context.pipeline_state.child_pipelines:
-            pause_action = TaskActionRequest(id=child_pipeline.id, action=TaskAction.UNPAUSE_PIPELINE)
-            context.send_message(context, child_pipeline.address, pause_action.id, pause_action)
+            # tell any child pipelines to resume
+            for child_pipeline in context.pipeline_state.child_pipelines:
+                pause_action = TaskActionRequest(id=child_pipeline.id, uid=_gen_id(), action=TaskAction.UNPAUSE_PIPELINE)
+                context.send_message(child_pipeline.address, pause_action.id, pause_action)
 
-    def cancel(self, context: TaskContext):
+        except Exception as ex:
+           # abort the pipeline if we could not resume the tasks
+            await self.cancel(context, ex)
+
+    async def cancel(self, context: TaskContext, ex=None):
         if context.pipeline_state.status.value not in [TaskStatus.PENDING, TaskStatus.RUNNING, TaskStatus.PAUSED]:
             raise ValueError(f'Pipeline is not in a state that can be cancelled')
 
@@ -123,14 +131,15 @@ class _Pipeline(object):
 
         # tell any child pipelines to cancel
         for child_pipeline in context.pipeline_state.child_pipelines:
-            cancel_action = TaskActionRequest(id=child_pipeline.id, action=TaskAction.CANCEL_PIPELINE)
+            cancel_action = TaskActionRequest(id=child_pipeline.id, uid=_gen_id(), action=TaskAction.CANCEL_PIPELINE)
             context.send_message(child_pipeline.address, cancel_action.id, cancel_action)
 
         # construct the cancellation exception to send to caller of this pipeline
-        cancellation_ex = _create_task_exception(context.storage.task_request, TaskCancelledException('Pipeline was cancelled'), context.pipeline_state.last_task_state)
+        ex = ex or TaskCancelledException('Pipeline was cancelled')
+        cancellation_ex = _create_task_exception(context.storage.task_request, ex, context.pipeline_state.last_task_state)
 
         # we move from cancelling to cancelled either by submitting and/or waiting on the finally task...
-        finally_task = self._graph.try_get_finally_task(context.get_caller_id())
+        finally_task = self._graph.try_get_finally_task()
 
         if finally_task is not None:
             if self.get_result_before_finally(context) is None:
@@ -143,4 +152,13 @@ class _Pipeline(object):
 
         else:
             # ...or by sending cancellation to ourself if there is no finally task
+            cancellation_ex.invocation_id = context.pipeline_state.invocation_id
+            cancellation_ex.uid = _gen_id()
             context.send_message(context.pipeline_state.address, context.pipeline_state.id, cancellation_ex)
+
+    def reset(self, context: TaskContext):
+        if context.pipeline_state is not None:
+            if not context.pipeline_state.status.value in [TaskStatus.COMPLETED, TaskStatus.FAILED, TaskStatus.CANCELLED]:
+                raise PipelineInProgress('Pipelines must have finished before they can be re-run')
+            else:
+                context.pipeline_state = None
