@@ -2,6 +2,7 @@ from statefun_tasks.context import TaskContext
 from statefun_tasks.pipeline_impl.handlers import PipelineMessageHandler
 from statefun_tasks.messages_pb2 import TaskRequest, TaskResult, TaskException, TaskStatus, PipelineState, ChildPipeline, TaskInfo
 from statefun_tasks.serialisation import pack_any
+from statefun_tasks.utils import _gen_id
 from google.protobuf.any_pb2 import Any
 from typing import Union
 
@@ -15,21 +16,23 @@ class BeginPipelineHandler(PipelineMessageHandler):
             and isinstance(message, TaskRequest) \
                 and not self.graph.is_empty()
 
-    def handle_message(self, context: TaskContext, message: Union[TaskRequest, TaskResult, TaskException], pipeline: '_Pipeline', task_state):
+    async def handle_message(self, context: TaskContext, message: Union[TaskRequest, TaskResult, TaskException], pipeline: '_Pipeline', state: Any):
         invoking_task = message  # type: TaskRequest
 
         # ensure we pick up the correct caller id when task producing this pipeline is a retry
-        if context.task_state.original_caller_id == '':
+        task_state = context.task_state.by_uid[invoking_task.uid]
+        if task_state.original_caller_id == '':
             caller_id = context.get_caller_id()
         else:
-            caller_id = context.task_state.original_caller_id
+            caller_id = task_state.original_caller_id
 
         # 1. record all the continuations into a pipeline and save into state with caller id and address
         context.pipeline_state = PipelineState(id = context.get_task_id(), address = context.get_address())
         context.pipeline_state.status.value = TaskStatus.RUNNING
         context.pipeline_state.pipeline.CopyFrom(pipeline.to_proto())
         context.pipeline_state.is_fruitful = pipeline.is_fruitful
-        context.pipeline_state.task_state.CopyFrom(pack_any(self.serialiser.to_proto(task_state)))
+        context.pipeline_state.task_state.CopyFrom(pack_any(self.serialiser.to_proto(state)))
+        context.pipeline_state.invocation_id = _gen_id()
 
         if caller_id is not None:
             context.pipeline_state.caller_id = caller_id
@@ -53,7 +56,7 @@ class BeginPipelineHandler(PipelineMessageHandler):
                 task.request = self._serialiser.serialise_args_and_kwargs(([]), {})
 
         # 3. split into tasks to call now and those to defer if max parallelism is exceeded
-        self.submitter.submit_tasks(context, tasks, max_parallelism=max_parallelism)
+        await self.submitter.submit_tasks(context, tasks, max_parallelism=max_parallelism)
 
         # 4. break
         return False, message
@@ -65,6 +68,7 @@ class BeginPipelineHandler(PipelineMessageHandler):
 
         child_pipeline = ChildPipeline(
             id = context.pipeline_state.id,
+            invocation_id = context.pipeline_state.invocation_id,
             address = context.pipeline_state.address,
             root_id = context.pipeline_state.root_id,
             root_address = context.pipeline_state.root_address,
@@ -73,7 +77,7 @@ class BeginPipelineHandler(PipelineMessageHandler):
         )
         
         for task in self.graph.yield_tasks(): 
-            child_pipeline.tasks.append(TaskInfo(task_id=task.task_id, task_type=task.task_type, namespace=task.namespace, worker_name=task.worker_name))
+            child_pipeline.tasks.append(TaskInfo(task_id=task.task_id, task_uid=task.uid, task_type=task.task_type, namespace=task.namespace, worker_name=task.worker_name))
 
         # notify back to the root pipeline
         context.send_message(child_pipeline.root_address, child_pipeline.root_id, child_pipeline)
