@@ -1,7 +1,7 @@
 from unittest.case import skip
 from statefun_tasks.context import TaskContext
 from statefun_tasks.pipeline_impl.handlers import PipelineMessageHandler
-from statefun_tasks.types import Task, Group
+from statefun_tasks.types import Task, Group, TasksException
 from statefun_tasks.messages_pb2 import TaskRequest, TaskResult, TaskException, TaskStatus
 from typing import Union
 
@@ -12,33 +12,49 @@ class ContinuePipelineHandler(PipelineMessageHandler):
     
     def can_handle_message(self, context: TaskContext, message: Union[TaskRequest, TaskResult, TaskException]) -> bool:
         return context.pipeline_state is not None \
-            and context.pipeline_state.status.value in [TaskStatus.RUNNING, TaskStatus.PAUSED] \
-                and isinstance(message, (TaskResult, TaskException))
+            and context.pipeline_state.invocation_id == message.invocation_id \
+                and context.pipeline_state.status.value in [TaskStatus.RUNNING, TaskStatus.PAUSED] \
+                    and isinstance(message, (TaskResult, TaskException))
 
-    def handle_message(self, context: TaskContext, message: Union[TaskRequest, TaskResult, TaskException], pipeline, **kwargs):
+    async def handle_message(self, context: TaskContext, message: Union[TaskRequest, TaskResult, TaskException], pipeline, **kwargs):
         task_result_or_exception = message
-        caller_id = context.get_caller_id()
+        caller_uid = message.uid
 
         # mark pipeline step as complete
-        self.graph.mark_task_complete(caller_id, task_result_or_exception)
+        self.graph.mark_task_complete(caller_uid, task_result_or_exception)
+
+        # notify event handler (with option to cancel)
+        try:
+            pipeline.events.notify_pipeline_task_finished(context, task_result_or_exception)
+        except TasksException as ex:
+            await pipeline.cancel(context, ex)
+            return False, task_result_or_exception
 
         # release deferred tasks if they can now run
-        self.submitter.release_tasks(context, caller_id, task_result_or_exception)
+        await self.submitter.release_tasks(context, caller_uid, task_result_or_exception)
 
         # get the next step of the pipeline to run (if any)
-        _, next_step, group, empty_group = self.graph.get_next_step_in_pipeline(caller_id)
+        current_step, next_step, group, empty_group = self.graph.get_next_step_in_pipeline(caller_uid)
 
         # if this task is part group then we need to record the results so we can aggregate later
         if group is not None:
-            self.result_aggregator.add_result(context, caller_id, task_result_or_exception)
+            await self.result_aggregator.add_result(context, caller_uid, task_result_or_exception)
 
             # once the group is complete aggregate the results
             if group.is_complete():
-                task_result_or_exception = self.result_aggregator.aggregate(context, group)
+                task_result_or_exception = await self.result_aggregator.aggregate(context, group)
+
+                # pause the pipeline if this completed group is a wait
+                if group.is_wait:
+                    await pipeline.pause(context)
+        else:
+            # pause the pipeline if this task is a wait
+            if current_step.is_wait:  
+                await pipeline.pause(context)
 
         # if we got an exception then the next step is the finally_task if there is one (or none otherwise)
         if isinstance(task_result_or_exception, TaskException):
-            next_step = self.graph.try_get_finally_task(caller_id)
+            next_step = self.graph.try_get_finally_task(caller_uid)
 
         # else if we came across an empty group between this task and next_entry then our result must be an empty array []
         # as we cannot call an empty group but can synthesise the result (remembering to pass through state)
@@ -62,7 +78,7 @@ class ContinuePipelineHandler(PipelineMessageHandler):
 
         if any(tasks):
             # split into tasks to call now and those to defer if max parallelism is exceeded
-            self.submitter.submit_tasks(context, tasks, task_result_or_exception=task_result_or_exception, max_parallelism=max_parallelism)
+            await self.submitter.submit_tasks(context, tasks, task_result_or_exception=task_result_or_exception, max_parallelism=max_parallelism)
 
         else:
             last_step = self._pipeline[-1]
