@@ -33,6 +33,19 @@ class PipelineGraph(object):
             else:
                 yield task_or_group
 
+    def is_top_level_group(self, group):
+        for task_or_group in self._pipeline:
+            if isinstance(task_or_group, Group) and task_or_group.group_id == group.group_id:
+                return True
+        return False
+
+    def get_task(self, task_id):
+        for task in self.yield_tasks():
+            if task_id == task.uid:
+                return task
+
+        raise ValueError(f'Task {task_id} not found')
+
     def is_empty(self):
         return not any(self.get_initial_tasks())
         
@@ -41,11 +54,11 @@ class PipelineGraph(object):
         def extract_task_chain(state, pipeline, entry):
             tasks = state
 
-            if entry.task_id == task_id:
+            if entry.uid == task_id:
                 matched = False
 
                 for task in self.yield_tasks(pipeline):
-                    if matched or task.task_id == task_id:
+                    if matched or task.uid == task_id:
                         matched = True
                         tasks.append(task)
 
@@ -102,13 +115,13 @@ class PipelineGraph(object):
         elif any(tasks):
             tasks[0].mark_complete()
          
-    def try_get_finally_task(self, task_id):
+    def try_get_finally_task(self, task_id=None):
         finally_task = next((task for task in self._pipeline if isinstance(task, Task) and task.is_finally), None)
-        return None if finally_task is None or finally_task.task_id == task_id else finally_task
+        return None if finally_task is None or finally_task.uid == task_id else finally_task
 
     def is_finally_task(self, task_id):
         finally_task = next((task for task in self._pipeline if isinstance(task, Task) and task.is_finally), None)
-        return finally_task is not None and finally_task.task_id == task_id
+        return finally_task is not None and finally_task.uid == task_id
 
     def get_next_step_in_pipeline(self, task_id):
         stack = deque([(self._pipeline, None, None, None, None)])  # FIFO (pipeline, entry after group, group, parent_group, entry after parent group)
@@ -133,7 +146,7 @@ class PipelineGraph(object):
                         stack.append((pipeline_in_group, stack_entry_after_group, stack_group_entry, stack_parent_group, stack_entry_after_parent_group))
 
                 else:
-                    if entry.task_id == task_id:
+                    if entry.uid == task_id:
                         next_entry = _try_next(iterator)
 
                         if next_entry is None and parent_group is not None and parent_group.is_complete():
