@@ -1,7 +1,8 @@
-from statefun_tasks import events
 from statefun_tasks.context import TaskContext
 from statefun_tasks.task_impl.handlers import MessageHandler
-from statefun_tasks.types import TASK_REQUEST_TYPE, TaskAlreadyExistsException
+from statefun_tasks.types import TASK_REQUEST_TYPE, PipelineInProgress
+from statefun_tasks.type_helpers import _create_task_exception
+from statefun_tasks.messages_pb2 import TaskRequest
 from statefun import Message
 from datetime import timedelta
 
@@ -15,17 +16,13 @@ class TaskRequestHandler(MessageHandler):
             task_input = message.as_type(TASK_REQUEST_TYPE)
             
             context.task_name = task_input.type
-            context.apply_task_meta(task_input)
+            context.contextualise_from(task_input)
 
             return task_input
         
         return None
 
-    async def handle_message(self, tasks: 'FlinkTasks', context: TaskContext, task_request):
-        if context.storage.task_request is not None:
-            # don't allow tasks to be overwritten
-            raise TaskAlreadyExistsException(f'Task already exists: {task_request.id}')
-
+    async def handle_message(self, tasks: 'FlinkTasks', context: TaskContext, task_request: TaskRequest):
         context.storage.task_request = task_request
 
         flink_task = tasks.get_task(task_request.type)
@@ -33,20 +30,23 @@ class TaskRequestHandler(MessageHandler):
         # notify started
         tasks.events.notify_task_started(context, task_request)
         
-        task_result, task_exception, pipeline, task_state = await flink_task.run(context, task_request)
-        
+        # run task code
+        task_result, task_exception, pipeline, state = await flink_task.run(context, task_request)
+
+        # if a pipeline is retured then try to reset its state in case it already exists
+        if pipeline is not None:
+            try:
+                pipeline.reset(context)
+            except PipelineInProgress as ex:  # can't re-run a running pipeline.  Must wait for it to finish
+                task_exception, pipeline = _create_task_exception(task_request, ex), None
+
         # notify finished
-        tasks.events.notify_task_finished(context, task_result, task_exception)
+        tasks.events.notify_task_finished(context, task_result, task_exception, is_pipeline=pipeline is not None)
 
         # if task returns a pipeline then start it if we can
-        if pipeline is not None and pipeline.handle_message(context, task_request, task_state):
+        if pipeline is not None and await pipeline.handle_message(context, task_request, state):
             ()
 
-        # else if we have a task result return it
-        elif task_result is not None:
-            context.storage.task_result = task_result
-            tasks.emit_result(context, task_request, task_result)
-
         # else if we have an task exception, attempt retry or return the error
         elif task_exception is not None:
             if self._attempt_retry(context, tasks, task_request, task_exception):
@@ -55,25 +55,34 @@ class TaskRequestHandler(MessageHandler):
             context.storage.task_exception = task_exception
             tasks.emit_result(context, task_request, task_exception)
 
+        # else if we have a task result return it
+        elif task_result is not None:
+            context.storage.task_result = task_result
+            tasks.emit_result(context, task_request, task_result)
+
     def _attempt_retry(self, context, tasks, task_request, task_exception):
+        task_state = context.task_state.by_uid[task_request.uid]
+
         if task_exception.maybe_retry and task_exception.retry_policy is not None:           
-            if context.task_state.retry_count >= task_exception.retry_policy.max_retries:
+            if task_state.retry_count >= task_exception.retry_policy.max_retries:
                 return False
 
-            # save the original caller address and id for this task_request prior to calling back to ourselves which would overwrite
-            if context.task_state.original_caller_id == '':
-                context.task_state.original_caller_id = context.get_caller_id()
-                context.task_state.original_caller_address = context.get_caller_address()
+            # increment retry count
+            task_state.retry_count += 1
 
-            # remove previous task_request from state, increment retry count
+            # save the original caller address and id for this task_request prior to calling back to ourselves which would overwrite
+            if task_state.original_caller_id == '':
+                task_state.original_caller_id = context.get_caller_id()
+                task_state.original_caller_address = context.get_caller_address()
+
+            # remove previous task_request from state
             del context.storage.task_request
-            context.task_state.retry_count += 1
 
             # send retry
             delay = timedelta(milliseconds=task_exception.retry_policy.delay_ms)
 
             if task_exception.retry_policy.exponential_back_off:
-                delay = timedelta(milliseconds=task_exception.retry_policy.delay_ms * (2 ** context.task_state.retry_count))
+                delay = timedelta(milliseconds=task_exception.retry_policy.delay_ms * (2 ** task_state.retry_count))
 
             if delay:
                 context.send_message_after(delay, context.get_address(), context.get_task_id(), task_request)
@@ -81,7 +90,7 @@ class TaskRequestHandler(MessageHandler):
                 context.send_message(context.get_address(), context.get_task_id(), task_request)
 
             # notify retry
-            tasks.events.notify_task_retry(context, task_request, context.task_state.retry_count)
+            tasks.events.notify_task_retry(context, task_request, task_state.retry_count)
 
             return True
 
