@@ -4,11 +4,14 @@ from statefun_tasks.pipeline_builder import PipelineBuilder
 from statefun_tasks.utils import _is_tuple, _type_name, _annotated_protos_for
 from statefun_tasks.messages_pb2 import TaskRequest
 from statefun_tasks.type_helpers import _create_task_result, _create_task_exception
+from statefun_tasks.types import YieldTaskInvocation
 
+from typing import AsyncGenerator, Generator
 import inspect
 import asyncio
 
 
+
 class FlinkTask(object):
     def __init__(self, fun, serialiser, events:EventHandlers, retry_policy=None, with_state=False, is_fruitful=True, with_context=False, **kwargs):
         self._fun = fun
@@ -38,15 +41,27 @@ class FlinkTask(object):
         # run the flink task
         try:
             fn_result = self._fun(*task_args, **kwargs)
+            
+            # convert generators to lists
+            if isinstance(fn_result, AsyncGenerator):
+                fn_result = list([x async for x in fn_result])
 
-            # await coro
-            if asyncio.iscoroutine(fn_result):
+            elif isinstance(fn_result, Generator):
+                fn_result = list([x for x in fn_result])
+
+            # await coros
+            elif asyncio.iscoroutine(fn_result):
                 fn_result = await fn_result
 
             pipeline, task_result, fn_state = self._to_pipeline_or_task_result(task_request, fn_result, fn_state)
 
-        # we errored so return a task_exception instead
+        
+        except YieldTaskInvocation as e:
+            # task yielded so we don't output anything
+            ()
+
         except Exception as e:
+            # we errored so return a task_exception instead
             task_exception = self._to_task_exception(task_request, e)
 
         return task_result, task_exception, pipeline, fn_state
@@ -87,6 +102,7 @@ class FlinkTask(object):
             fn_result = ()
             
         task_result = _create_task_result(task_request)
+
         self._serialiser.serialise_result(task_result, fn_result, fn_state)
 
         return pipeline, task_result, fn_state
