@@ -1,4 +1,4 @@
-from statefun_tasks.utils import _type_name
+from statefun_tasks.utils import _type_name, _gen_id
 from statefun_tasks.messages_pb2 import PipelineState, TaskState, TaskRequest, TaskResult, TaskException, \
     TaskActionRequest, TaskActionResult, TaskActionException, TaskEntry, GroupEntry, PipelineEntry, TaskRetryPolicy, \
     Pipeline, ChildPipeline
@@ -54,7 +54,7 @@ class Task:
 
     @staticmethod
     def from_fields(task_id, task_type, task_args, task_kwargs, is_finally=None, namespace=None, worker_name=None,
-                    is_fruitful=None, retry_policy=None, display_name=None, **kwargs):
+                    is_fruitful=None, retry_policy=None, display_name=None, is_wait=None, uid=None, **kwargs):
         proto = TaskEntry(
             task_id=task_id,
             task_type=task_type,
@@ -64,10 +64,19 @@ class Task:
             worker_name=worker_name,
             is_fruitful=is_fruitful,
             retry_policy=retry_policy,
-            display_name=display_name
+            display_name=display_name,
+            is_wait=is_wait,
+            uid=uid if uid is not None else _gen_id()
         )
         return Task(proto, task_args, task_kwargs)
-        
+
+    @property
+    def uid(self):
+        """
+        The unique ID of this task
+        """
+        return self._proto.uid
+
     @property
     def id(self):
         """
@@ -82,6 +91,10 @@ class Task:
         """
         return self._proto.task_id
 
+    @task_id.setter
+    def task_id(self, value):
+        self._proto.task_id = value
+
     @property
     def task_type(self):
         """
@@ -166,6 +179,14 @@ class Task:
         else:
             self._proto.display_name = value
 
+    @property
+    def is_wait(self):
+        return self._proto.is_wait
+
+    @is_wait.setter
+    def is_wait(self, value):
+        self._proto.is_wait = value
+
     def unpack(self, serialiser):
         if self._unpacked or not self._proto_backed:
             return self
@@ -209,9 +230,10 @@ class Task:
 
 
 class Group:
-    def __init__(self, group_id, max_parallelism=None):
+    def __init__(self, group_id, max_parallelism=None, is_wait=None):
         self.group_id = group_id
         self.max_parallelism = max_parallelism
+        self.is_wait = is_wait
         self._group = []
 
     def add_to_group(self, tasks):
@@ -230,7 +252,7 @@ class Group:
         return None  # _GroupEntries don't have a single destination
 
     def to_proto(self, serialiser) -> PipelineEntry:
-        proto = GroupEntry(group_id=self.group_id, max_parallelism=self.max_parallelism)
+        proto = GroupEntry(group_id=self.group_id, max_parallelism=self.max_parallelism, is_wait=self.is_wait)
 
         for entries in self._group:
             pipeline = Pipeline()
@@ -245,7 +267,7 @@ class Group:
 
     @staticmethod
     def from_proto(proto: PipelineEntry):
-        entry = Group(group_id=proto.group_entry.group_id, max_parallelism=proto.group_entry.max_parallelism)
+        entry = Group(group_id=proto.group_entry.group_id, max_parallelism=proto.group_entry.max_parallelism, is_wait=proto.group_entry.is_wait)
 
         group = []
         for pipeline in proto.group_entry.group:
@@ -281,11 +303,26 @@ class RetryPolicy:
             exponential_back_off=self.exponential_back_off)
 
 
-class TaskAlreadyExistsException(Exception):
+class TasksException(Exception):
     def __init__(self, message):
         super().__init__(message)
 
 
-class TaskCancelledException(Exception):
+class TaskAlreadyExistsException(TasksException):
     def __init__(self, message):
-        super().__init__(message)
\ No newline at end of file
+        super().__init__(message)
+
+
+class TaskCancelledException(TasksException):
+    def __init__(self, message):
+        super().__init__(message)
+
+
+class PipelineInProgress(TasksException):
+    def __init__(self, message):
+        super().__init__(message)
+
+
+class YieldTaskInvocation(Exception):
+    def __init__(self):
+        super().__init__()
